<?php


// $Id: administer_predicates.inc,v 1.41 2006/10/14 15:52:58 dman Exp $

/**
 * @file
 *
 * Admin functions to administer the predicate terms definitions.
 *
 * Functions here relate only to the admin screens where
 * we can add and edit term definitions. None of the functions
 * in this file should be required to be loaded in a read-only
 * environment.
 *
 * Terms can be defined and added to the term table,  preferably via
 * import from RDF.
 * 
 * To get a new predicate to show up in the available relationship types, (left
 * drop down of the node edit) add the 'node type' as a member of the
 * <em>domain</em> of the desired predicate or edit the definition of the node
 * type class_def_link to add the predicate as an <em>available property</em>.
 * Saying 'this property can apply to these types of content' or 'this type of
 * content can have this property' are equivalent.
 * 
 *
 * Pluggable Predicate Ontologies
 *
 * A module such as ontology_family may be created and enabled. It would
 * declare a bunch of predicates 
 * (IsFatherOf,  IsMotherOf,  IsParentOf, IsChildOf)
 * and their characteristics ( 
 *  IsMotherOf isSubPropertyOf IsParentOf,  
 *  IsChildOf isInverseOf IsParentOf
 *)
 * These predicates are added to the term list when the module is enabled,
 * and become available (on node edit) to anyone who wishes to use them. 
 * 
 * Ontologies are best defined and distributed in RDF files.
 *
 * To restrict or attach certain ontologies to only certain TYPES of node (for
 * example, you may create a custom node type for 'person', and only persons
 * have family relations) the 'domain' of a term can be set. This is analogous
 * to associating node types with a vocabulary.
 * 
 * When sets of terms are grouped in ontologies, node types can be associated
 * with entire ontologies, which is a tidier way of applying related properties.
 *
 * Characteristics
 *
 * ... are axiomatic (basic,assumed) properties of predicates.
 *
 * Available characteristics should be used to relate one term with another.
 * Characteristics are modelled on the OWL definitions,  and include some
 * basic 'meanings' of term inter-relations. @see apply_characteristics.inc
 *
 * But sometimes that's not enough. 
 *
 * Custom query responses
 *
 * Predicate Ontology Modules can also define predicate logic lookups,
 * such as a routine that is called to calculate the results of a request to
 * 'IsCousinOf'.
 * This would return a meaningful result when the the system is asked for all
 * the cousins. This sort of lookup requres some coded,  repetative graph 
 * walking. If you wanted to encode the logic of a lookup HeirToTheThrone - 
 * this is where it would be done.
 * 
 * The functions are hooks of the form relationship_lookup_hook() where 'hook'
 * is the name of the predicate in question.
 *
 * These graph walks may be arbitrarily complex, and may need to call in 
 * many other inferred statements to get to their result. This on-demand
 * sort of query is different from the 'figure out what it means when we get it'
 * or on-save inferencing.
 * 
 * Internally, an example is relationship_lookup_CanHaveProperty(). This
 * callback is fired every time a query is made for the 'CanHaveProperty'
 * relationship. It's a callback, because CanHaveProperty requires class
 * inheiritance support, and the process to answer that question requires the
 * code to look up all owner classes, the properties THEY may have, recursively.
 *
 * Custom relationship renderings
 *
 * The ontology libraries are also expected to provide appropriate ways of
 * displaying the relationships at node view time,  whether this be by publishing
 * a 'see also' sort of navigation block,  @see ontology_navigation.module 
 * or using hook_nodeapi() to modify the body output. 
 * @see ontology_revision_control.module
 * to see how isReplacedBy causes a warning to appear at the top of a page 
 * that is outdated.
 *
 * See the ontology modules for examples on how to make up
 * your own predicate programatically.
 *
 * 
 * Modules can also define their own characteristics and the logical
 * interpretation of what they imply,  however that is tricky and should be
 * rare.
 *
 * @link predicates See Predicates for more about how I deal with predicate terms. @endlink
 * @see apply_characteristics.inc for how to code a characteristic.
 *
 * @todo Use namespaces internally for indexing predicate terms? 
 *
 * @see ontologies
 * 
 * @ingroup docs
 * @package coders
 * @author Dan Morrison http://coders.co.nz/
 */

/**
 * Menu callback;
 * @ingroup relationship_admin
 */
function relationship_admin(){
  $output = relationship_overview_page();
  print theme('page', $output);
}

/**
 * List a summary of all terms and the actions we can do on them.
 * @returns html table
 * @ingroup relationship_admin
 */
function relationship_overview_page(){
  $op = $_REQUEST['op'];
  $edit = $_POST['edit'];

  if ($op == t('Change Preferences')){
    foreach ($edit as $param => $val){
      variable_set($param, $val);
    }
  }

  if ($op == t('Hide Advanced Ontologies')){
    variable_set('relationship_show_advanced_terms', FALSE);
  }
  if ($op == t('Show Advanced Ontologies')){
    variable_set('relationship_show_advanced_terms', TRUE);
  }

  $header = array (
    t('Name'), 
    t('Label'), 
    t('Characteristics'), 
    array ('data' => t('Operations'), 'colspan' => '3')
  );

  // Start with all known predicates, so we know which are left over as unsorted
  $terms = relationship_get_terms();
  # debug_pre(array_keys($terms),0);

  // We'll list anything that's in an ontology
  $show_advanced = variable_get('relationship_show_advanced_terms', FALSE);
  // but filter some ontologies off the UI for convenience
  $hidden_ontologies = variable_get('relationship_hide_ontologies',array());
  

  // First enumerate the available, defined ontologies
  // TODO - do ontologies really need ordering?
  $ontologies = relationship_get_resources_matching_query(PREDICATE_TYPE, CLASS_ONTOLOGY);
  debug("Enumerating known ontologies :".join(", ",array_keys($ontologies)),1);
  foreach ($ontologies as $ontology_name => $ontology_def){
    // Skip the confusing technical ones unless admin has enabled that
    if ((!$show_advanced) && ( $hidden_ontologies[$ontology_name])){
      debug("Not listing members of $ontology_name, it's hidden by request ".join(', ',array_values($hidden_ontologies)) ,2); 
      $skip = TRUE;
      // even when skipping we need to iterate to cross these off the list.
      $hidden_ontologies[$ontology_name] = relationship_get_display_label($ontology_def);
    } 
    else {
      $skip = FALSE;
      unset($hidden_ontologies[$ontology_name]); // just for formatting the note below
      
      // prepare to display the block
      $data = relationship_get_display_label($ontology_def);
      $actions  = l(t('edit ontology'), relationship_term_path($ontology_def,'edit') );
      $rows[] = array (
        array ( 'data' => '&nbsp;', 'colspan' => '5' )
      );
      $rows[] = array (
        array ( 'data' => $data, 'colspan' => '3',  'style' => 'font-weight:bold' ),
        array ( 'data' => $actions )
      );
    }
      
   
    // Now retrieve all terms in this ontology, and format them

    // A loaded ontology should already know what its parts are.
    // however, this lookup also does the trick:
    //  $terms_in_ontology = relationship_get_resources_matching_query(PREDICATE_IS_PART_OF, $ontology_def->uri,TRUE);

    $terms_in_ontology = array();
    if($ontology_def->hasPart){
      // ensure the names are cannonic, as a mix of long and shortnames produces dupes.
      $ontology_def->hasPart = array_unique( relationship_get_longname($ontology_def->hasPart) );
      foreach((array)$ontology_def->hasPart as $part_name ){
        // try the term cache (for efficiency) first
        if($term = relationship_get_term($part_name)){
           $terms_in_ontology[$part_name] = $term;
        }
      }
    }
    debug("Listing terms defined in ontology '$ontology_name' " . print_r($ontology_def->hasPart,1) . ($skip ? "(hidden)" : ""), 2);

    foreach ($terms_in_ontology as $term){
      unset ($terms[$term->name]); // Tick this off as done
      if($skip){continue;}
      if( ($term->type != CLASS_PROPERTY) && (!$show_advanced) ){continue;}
      debug("Listing Term '$term->name' under $ontology_name", 3);
      $rows[] = _format_term_row($term);
    }
    if(empty($terms_in_ontology)){
      $rows[] = array(array('data' => '[none]', 'colspan' => '5'));
    }
  }

  // Possible (but rare) that one of the original terms was not sorted into an ontology. 
  // List it anyway, so we can fix that. It'd be hard to find otherwise
  if ($terms){
    $rows[] = array ( array ( 'data' => '&nbsp;', 'colspan' => '5' ) );
    $rows[] = array ( array ( 'data' => 'Unsorted', 'colspan' => '5', 'style' => 'font-weight:bold' ) );
    foreach ($terms as $term){
      if (!$term){ continue; }
      $rows[] = _format_term_row($term);
    }
  }

  if (!$rows){
    $query = "op=Import&edit[source]=" . drupal_get_path('module','relationship') . '/axioms.rdf' ;
    $rows[] = array ( array ( 'data' => t('No terms available.'.l("Load Axioms", "admin/relationship/tidy", NULL, $query)), 'colspan' => '5' ));
  }

  $output = theme('table', $header, $rows);

  if (variable_get('relationship_show_advanced_terms', FALSE)){
    $prefs .= t("
                <p>
                  The terms and classes shown here are all resources that are either of 
                  <strong>type owl:Property</strong>  or anything that 
                  <strong>isPartOf</strong> any groupings of <strong>type owl:Ontology</strong>.
                </p><p>
                  The grouping displayed here is a result of a terms <em>membership</em>
                  in local <em>ontologies</em>. If 'domain' 'isPartOf' 'rdfOntology'
                  it will be grouped there. Terms can be members of multiple ontologies.
                  The <em>visibility</em> of ontologies on this page can be adjusted in the <a href='%settings_link'>settings</a>.
                  <br/>
                  Ontologies are useful to define a whole family of related terms,
                  and that ontology can then be associated with a content type 
                  (see either the ontology or <a href='%content_type_link'>content type configuration</a> pages)
                  to provide all the terms to that type at once.
                </p>
                ",
      array(
        '%settings_link' => url('admin/settings/relationship'),
        '%content_type_link' => url('admin/settings/content-types'),
      )
    );
    $button_action = t('Hide Advanced Ontologies');
  }
  else{
    $prefs .= t("
                <p>
                  Some technical terms (%hidden_list) and the Ontologies and Classes
                  may be currently <i>hidden</i> on this screen. 
                </p>
                ",
      array('%hidden_list' => join(', ',$hidden_ontologies) )
    );
    $button_action = t('Show Advanced Ontologies');
  }
  $buttons = array(
    l($button_action, 'admin/relationship', array (), 'op=' . $button_action),
    l('More Settings', 'admin/settings/relationship')
  );
  $prefs .= theme('item_list',$buttons);

  $output .= $prefs;

  $output .= t("
  <p>
    The best way to define new terms is to install an ontology module which
    defines its own predicates, 
    absorb an additional ontology <a href='%database_link'>from rdf/XML files</a> 
    or install the <b>rdf_ontology.module</b> 
    and import globally-defined terms from the web using that.
    <br/>Plugin Ontology modules should be available from the Drupal module repository.
  </p>
  ",array('%database_link' => url('admin/relationship/tidy')));

  return $output;
}

/**
 * Format display of the defined term or characteristic properties.
 * Returns an array for use as a tr.
 * helper function for relationship_overview()
 */
function _format_term_row($term){
  $relationship = "";

  if(! relationship_term_is_valid($term)){
    drupal_set_message("Problem with term definition ".print_r($term,1),'error'); 
  }
  
  if(($term->type != CLASS_PROPERTY)){
    // show more info
    $relationship = $term->type;
  }

  $actions  = l(t('edit term'), relationship_term_path($term,'edit') );
  $actions .= ' '. l(t('delete'), relationship_term_path($term, 'delete') );

  return array (
    ($term->uri ? l($term->name, $term->uri, array('title' => 'Formal definition of this term') ) : $term->name),
    l($term->label, relationship_term_path($term,'view'), array ('title' => $term->description) ),
    //    '<acronym title="'.check_plain((string)$term->description).'" >'.check_plain((string)$term->label).'</acronym>',
    $relationship, 
    $actions
  );
}


/**
 * Menu callback; Edits or adds a predicate term
 * @ingroup relationship_admin
 */
function relationship_admin_term($op = 'add', $term = ''){
  # drupal_set_message("$op '$term'");

  if(!$term){$term = $_REQUEST['edit']['subject'];}
  if ($term){
    drupal_set_title("edit term $term");
    $o_term = relationship_get_term($term, TRUE);

    if ($op == t('delete')){
      relationship_delete_term($term); 
      drupal_set_message("Deleted '$term'");
      relationship_flush_terms();
      drupal_goto('admin/relationship');
    }

  }
  else{
    drupal_set_title('add term');
    $o_term = (object) array (
      'type' => CLASS_PROPERTY,
    );
  }
  $term_form = _relationship_predicate_form($o_term);
  $output = drupal_get_form('relationship_predicate_form', $term_form);
  // The name of this form also defines what happens when it's submitted

  print theme('page', $output);
}

/**
 * Routines for displaying the admin forms.
 */


/**
 * Display an edit form for a predicate 'term'.
 * 
 * Set its name,  label,  characteristics and descriptions here.
 * Should also work for 'new' terms.
 * @param $edit assoc array of edit fields
 * @return HTML form
 * @ingroup relationship_admin
 */
function _relationship_predicate_form($term_def = array ()) {
  $edit = (array) $term_def;
  $assist_scripts = "\n/* Little form usability additions */\n";
  #dsm($term_def);
  $form['label'] = array (
    '#type' => 'textfield',
    '#title' => $term_def->type .' '. t('label'),
    '#default_value' => $edit['label'],
    '#size' => 50,
    '#maxlength' => 64,
    '#description' => t("
                          The human-readable predicate for this resource. 
                          Use a descriptive phrase with leading caps. 
                          Choose one from <a href='http://dublincore.org/documents/dcmi-terms/' 
                          target='_new'>the Dublin Core list</a> if possible. 
                          For predicates, this should be a descriptive sentence fragment that will make grammatical
                          sense of the statement 'Document A [....] Document B'.  
                          Example: 'Is Part Of', 'References', 'Is Replaced By'.
                        "), 
    '#attributes' => NULL,
    '#required' => TRUE,
    );

  // NOTE, although the statement us about the identifier, the object has the property 'name'
  // they mean the same thing.
  $form['name'] = array (
    '#type' => 'textfield',
    '#title' => $term_def->type .' '.t('identifier'),
    '#default_value' => $edit['name'], 
    '#size' => 50, 
    '#maxlength' => 64,
    '#description' => t("
                          The internal name/id. 
                          May be the same as,  or a shorthand (CamelCase) form of the label. 
                          Should match the <a href='http://dublincore.org/documents/dcmi-terms/' 
                          target='_new'>Dublin Core naming convention</a> if possible. 
                          Example: 'isPartOf', 'References', 'IsReplacedBy'.
                        "), 
    '#attributes' => array ( 'onclick' => "update_name(this)" ), 
    '#required' => TRUE,
  );
  $assist_scripts .= "function update_name(element){if(!element.value){element.value = element.form.elements['edit-label'].value.replace(/ /g,'');}}\n";

  $form['description'] = array (
    '#type' => 'textarea',
    '#title' => t('Description'),
    '#default_value' => $edit['description'],
    '#cols' => 60,
    '#rows' => 5,
    '#description' => t("
                          A description of what relationship the term is supposed to describe.
                        "),
    '#attributes' => array ('onclick' => "update_description(this)"),
  );
  $assist_scripts .= "function update_description(element){if(!element.value){element.value = 'This resource '+element.form.elements['edit-label'].value+' the referenced resource.'}}\n";

  if ($term_def->derived)
  {
    $form[] = array (
      '#value' => t("
          This property is <em>derived</em> which means it is never
          stated explicitly by a user, but is set only through deduction
          based on other relationships. Exactly what those relationships 
          <em>are</em> is defined in the characteristics below.
        ")
    );
  }
  $form['uri'] = array (
    '#type' => 'textfield',
    '#title' => t('URI' ),
    '#default_value' => $edit['uri'], 
    '#size' => 50,
    '#maxlength' => 64, 
    '#description' => t("
                          A reference to the definative resource (schema) that describes this term.
                          A technical definition of what properties this metadata term implies. 
                          Eg: 
                            'http://purl.org/dc/terms/replaces', 
                            'http://www.w3.org/TR/html4/types.html#type-links' . 
                          Should probably be some implimentation of 
                          <a href='http://www.w3.org/TR/owl-features/' title='Web Ontology Language'>OWL</a>
                          which can then define term properties (characteristics) like 
                          'commutative' and 'associative'. 
                          Or you can just make your own URI up or leave it blank.
                          Even if its not a formal schema, it is still used as a namespace.
                          Schemas can be found at 
                          <a href='http://www.schemaweb.info' target='_new'>Schemaweb</a>
                          and other places.
                        "),
    '#attributes' => array ('onclick' => "update_uri(this)"),
    '#required' => FALSE,
  );
  $assist_scripts .= "function update_uri(element){if(!element.value){element.value = element.form.elements['edit-" . 'name' . "'].value;}}\n";

  $type_urls = relationship_query(NULL, PREDICATE_TYPE, CLASS_CLASS, 'subject');

  foreach ($type_urls as $url){
    $types[$url] = relationship_get_shortname($url);
  }
  // use longnames under the hood, display shortnames.
  $form['type'] = array (
    '#type' => 'select',
    '#title' => t('Type'), 
    '#default_value' => relationship_get_longname($edit['type']), 
    '#options' => $types, 
    '#description' => t("
                          The type of resource this is. 
                          Predicate Terms are of type <strong>Property</strong>
                          but advanced options are possible if OWL classes are being defined.
                          <em>Changing</em> this on the fly is probably a really bad idea.
                        "),
  );
  if(! user_access('edit intrinsic properties')){
    $form['type']['#attributes']['disabled'] = TRUE; // 4.7
    $form['type']['#disabled'] = TRUE; // Drupal 5
  }

/*
  $form['weight'] = array (
    '#type' => 'weight',
    '#title' => t('Weight' ), 
    '#default_value' => $edit['weight'],
    '#delta' => 10,
    '#description' => t(" Importance for display and listings. "),
    );
  $form['help'] = array (
    '#value' => t(" 
                  It's not always appropriate for all terms to show up on all screens,
                  What gets displayed where can be controlled with a
                  combination of accurate 'domain' settings, and the
                  use of ontology groupings.

                  Term 'nav:last' would be an <em>implied</em> property only (derived from
                  nav:next) and although it provides part of an ontology, it is not 
                  something the user should override. 
              "),
  );
*/

  $form['help'] = array (
    '#value' => t(" 
                  What term gets displayed where can be controlled with a
                  combination of accurate 'domain' settings, and the
                  use of ontology groupings.
              "),
  );


  // Use the node relationship form here to apply further statements to this form
  $characteristics = relationship_get_all_characteristics($term_def->type);

  $form[] = relationship_form($term_def,$characteristics);
  $form['#submit'] = array(
    'relationship_predicate_form_submit'=>array(),
  );


  $form['Submit'] = array (
    '#type' => 'submit',
    '#value' => t('Submit'),
  );

  if ($edit['name']){
    $form['Delete'] = array (
      '#type' => 'submit',
      '#value' => t('Delete'),
    );
  }

  $form['glossary'] = _relationship_format_glossary($characteristics);
  
  /* Advanced diagnostic dump */
  if ($edit && variable_get('relationship_show_advanced_terms', FALSE)){
    $dump = var_export($edit, 1);
    $form[] = array (
      '#value' => "<h4>'Advanced Terms' is enabled in the settings</h4>Dumping the structure for inspection. Turn off 'advanced' to stop this" 
    );
    $form[] = array (
      '#value' => "<div><textarea rows='" .
       (count(explode("\n",
      $dump
    )) + 2) . "' cols='80'>" . $dump . "</textarea></div>");
  }

  if ($edit['uri'] && function_exists('relationship_rdf_render_rdf_about')){ 
  // RDF interface is optional
    $form['RDF'] = _relationship_format_rdf_dump($edit);
  }

  drupal_set_html_head("<script type='text/javascript' >\n//<![CDATA[\n$assist_scripts\n//]]>\n</script>");

  return $form;
}

function relationship_predicate_form_validate($form_id, $form_values){
  debug("validating submitted ".$form_id,2);
}

/**
 * Accept the form submission for a term and save the result.
 */
function relationship_predicate_form_submit($form_id, $resource){
//  debug("Submitted a ".$form_id,0);
//  debug_pre($resource);
//  dsm(array($resource,$_REQUEST));

  $resource = (object) $resource;

  if ($_REQUEST['op'] == t('Delete')){
    drupal_set_message(t('Deleting relationship term "%name" ', array (
      '%name' => $resource->name
    )));
    return relationship_delete_term($resource->name);
  }

  relationship_read_form($resource);
  relationship_merge_statements_into_resource($resource);

  relationship_save_term($resource);
  relationship_flush_terms();

  return 'admin/relationship';
}




/**
 * Return a help block describing all the given terms and characteristics
 */
function _relationship_format_glossary($characteristics){

  $glossary = "<p>
                There may be relationship <em>between</em> these relationship,  eg 
                '<strong>References</strong>' is the <em>inverse</em> of 
                '<strong>Is Referenced By</strong>',  and a '<strong>Is Part Of</strong>'
                property is <em>transitive</em> because it's implied that a part of a 
                part also a part of the whole.
            </p><p>
                For more detail,  refer to the 
                <a href='http://www.w3.org/TR/2004/REC-owl-guide-20040210/#PropertyCharacteristics'>
                Web Ontology Language</a>. 
                If used,  this extra information can be used as rules for 'graph walking'
                to deduce implicit metadata for items without stating them explicitly 
                each time,  eg a page can infer which pages reference it without 
                maintaining a list of referees itself.
            </p>\n";

  foreach ($characteristics as $char_name => $char_def){
    $glossary .= t("\n<dt>
      <a href='%cannonic_link' title='See the Formal definition of this characteristic'>%identifier</a> : 
      <a href='%local_link' title='See the Local definition of this characteristic'>%label</a>
      </dt>\n<dd>{$char_def->description}</dd>\n",
      array(
        '%cannonic_link' => $char_def->uri ,
        '%identifier' => $char_def->name ,
        '%local_link' => url(relationship_term_path($char_def,'edit')),
        '%label' => $char_def->label ,
     )
    );
  }

  $glossary .= '</dl>';
  return array (
    '#type' => 'fieldset',
    '#title' => t('About Property Characteristics' ),
    '#description' => $glossary,
    '#collapsible' => TRUE,
    '#collapsed' => TRUE,
    );
}

function _relationship_format_RDF_dump($edit){
    $RDF = array ( 
      '#type' => 'fieldset',
      '#title' => 'RDF',
      '#collapsible' => TRUE, 
      '#collapsed' => TRUE, 
    );
    $RDF['title'] = array(
      '#value' => l('RDF', "relationship/rdf/" , 
        array ('title' => t('The pure RDF-metadata about this resource.') ),
        "edit[uri]=".urlencode( $edit['uri'] )
      )
    ); 
    // when dumping an ontology, dump all terms too
    if($term_def->type == CLASS_ONTOLOGY){
      $subjects =  (array)$term_def->hasPart;
      array_unshift( $subjects, $term_def->uri);
      # debug_pre($subjects,2);
    }
    
    $RDF['dump'] = array( '#value' =>              
      '<div style="width:500px; overflow:scroll">'
      . relationship_rdf_render_rdf_about( $subjects?$subjects:$term_def->uri)
      . '</div>'
    );  
    return $RDF;
}

/**
 * Return a link to this terms edit/property screen
 * @ingroup relationship
 */
function relationship_term_path($term,$action='view') {
  $action = $action ? $action.'/' : '';
  if(variable_get('relationship_long_names', FALSE))
    return "admin/relationship/term/$action?edit[subject]=" . urlencode($term->uri);
  return "admin/relationship/term/$action".($term->identifier?$term->identifier:$term->name);
}




/**
 * Display an summary of the properties of the given term
 * @param $term_def 
 * @return HTML
 * @ingroup relationship_admin
 */
function relationship_term_overview($term = array ()) {
  if(is_string($term)){
   $term = relationship_get_term($term); 
  }
  $out .= '<h3>'.$term->label.'</h3>';
  $out .= '<h4>Identifier: <code>'.$term->name.'</code></h4>';
  $out .= '<p>' .$term->description.'</p>';
  $out .= '<h4>'.$term->uri.'</h4>';
  $out .= '<pre>'.print_r($term,1).'</pre>';
  
  $domains = array();
  foreach((array)$term->domain as $domain){
    $t = relationship_get_term($domain);
    /* 4.8
    $item = array(
      'data' => l($item->label,  relationship_term_path($item,'edit') )
    );
    */
    /* 4.7 */
    $item = l($domain,  relationship_term_path($t,'edit') );
    $domains[$domain] = $item;
  }
  if(!$domains){$domains=array('Unrestricted domain - all resources');}
  $domain_list = theme_item_list($domains);
  
  foreach((array)$term->range as $range){
    $t = relationship_get_term($range);
    $item = l($domain,  relationship_term_path($t,'edit') );
    $ranges[$range] = $item;
  }
  if(!$ranges){$ranges=array('Unrestricted range - any resource');}
  $range_list = theme_item_list($ranges);

  $header = array(
    t('Resource types that can <em>have</em> this type of relationship. (<em>subjects</em>)'),
    t('Resource types that may be <em>targets</em> of this type of relationship. (<em>objects</em>)'),
  );
  $out .= theme_table($header, array(array($domain_list,$range_list)));
  
  return $out;
}


/**
 * Save a fully-defined term.
 * 
 * This does NOT create statements from values, it just serializes the
 * statements already present. 
 * @see relationship_predicate_form_submit()
 * 
 * Deal with a few special cases,  then serialize the data
 * @param $edit assoc array of edit fields
 * @return assoc array,  possibly slightly modified for the serialization
 * @ingroup relationship_admin
 */
function relationship_save_term($edit) {
  $term = (object) $edit;

  $term->type = relationship_get_shortname($term->type);

  // Special-case the 'Inverse' property
  // tell the other relationship to point back to us
  if ($other_predicate_assertion = $term->statements['inverseOf']){
    $other_predicate = $other_predicate_assertion['object'];
    drupal_set_message("As it's an inverse,  I'm linking target term $other_predicate back to this one as $characteristic also.");
    $other_term = relationship_get_term($other_predicate);
    $other_term->characteristics['inverseOf'] = $term->name;
    $other_term->inverseOf = $term->name;
  }
  
  if ($term->name){
    $term->statements = array_merge($term->statements,relationship_array2statements( $term ) );
    relationship_update_resource_statements($term);
    drupal_set_message( t("The term %term has been saved.", 
      array ('%term' => theme('placeholder', $term->name ) )
    ));
  }
  else{
    drupal_set_message( t("Unnamed term was not updated") );
    debug_pre($term);
  }
  // remember to relationship_flush_terms(); if needed

  return $term;
}

/**
 * @ingroup relationship_admin
 */
function relationship_delete_term($name, $extreme_prejudice=FALSE) {
  if(!$name){debug('Invalid argument ('.$name.') to relationship_delete_term()',1);return;}
  debug("Deleting the term '$name'",2);

  $uri = relationship_get_longname($name);
  relationship_resource_delete_statements($uri);
  
  if($extreme_prejudice){
    $ds = _relationship_datastore();
    drupal_set_message("Deleting all statements using the term '$name'",2);
    $ds->delete(NULL,$name,NULL) ;
    $ds->delete(NULL,$uri,NULL) ;
  }
  // remember to flush after yourself, this term may have left smears in the cache
}


/**
 * @}
 */



/**
 * Performs a consistancy check on the database.
 *
 * Removes references to non-existant nodes
 * Allows a refresh of the predicate data.
 * Menu callback
 * @ingroup relationship_admin
 */
function relationship_check_database(){
  $op = $_REQUEST['op'];
  $edit = $_REQUEST['edit'];

  /* Clean up nids that don't exist */

  if ($op == t('Drop')){
    drupal_set_message('Dropping ALL data from database');
    db_query('DELETE FROM {relationship_statement}');
    relationship_flush_terms();
  }

  $out = "<p>
          <h2>Cleanup database</h2>
          Although it should not happen, it's possible that we could be left
          with references concerning nodes that no longer exist. This page can 
          check and clean these up. It's really just an admin function for
          development.
          </p>
  		  ";

  // statement triples cannot be empty or null
  $invalid_statement_query = " (TRIM(subject) = '') OR (TRIM(predicate) = '') OR (TRIM(object) = '') ";
  $query = "SELECT s.%s as nid FROM {relationship_statement} s LEFT JOIN {node} n ON ( s.%s = n.nid) WHERE s.%s > 0 AND n.nid IS NULL GROUP BY s.%s ";
  $misdirected_subjects = db_query($query, 'subject', 'subject', 'subject', 'subject');
  $out .= '<p>Found ' . db_num_rows($misdirected_subjects) . " statements in DB with an invalid subject<br/>";

  $form = array ();
  while ($s = db_fetch_object($misdirected_subjects)){
    $out .= 'node:' . $s->nid . '<br/>';
    $form['nid'][''] = array (
      '#type' => 'hidden',
      '#value' => $s->subject,
    );
  }
  $out .= '</p>';

  $query = "SELECT rsid FROM {relationship_statement} WHERE $invalid_statement_query";
  $invalid_statements = db_query($query);
  $out .= '<p>Found ' . db_num_rows($invalid_statements) . " statements in DB that have invalid options<br/>";

  $out .= '</p>';
  $form['nid'][''] = array (
    '#type' => 'submit',
    '#value' => t('Clean Up'),
  );

  $out .= drupal_get_form('relationship_database_maintainence', $form);

  /* Check remote URLs? */

  $out .= ' ' . l(t('Drop'), 'admin/relationship/tidy', array (), 'op=' . t('Drop'));

  $out .= "<p>
  		      <h2>Reset Axiomatic Predicates</h2>
  		      In the database, there are a set of underlying 'axioms' defining
  		      The very definition of the terms used as predicates. 
  		      These sometimes need tweaking. Use the below import buttons to
            make sure they are up-to-date.
     		  </p>";



  // Importing from Ontologies
  $import_command = "Import";
  $schema_files = file_scan_directory( dirname(drupal_get_filename('module','relationship')),"^[^\.].*\.rdf",array('.', '..', 'CVS'),NULL,TRUE );
  foreach($schema_files as $i => $schema_file){
    $imports[$i] = "Import ".l($schema_file->name, 'admin/relationship/tidy', NULL, "op=$import_command&edit[source]=".$schema_file->filename );
  }
  $out .= theme_item_list($imports,"Import Additional Ontologies");

  if($op == $import_command){
    $source = $edit['source'];
    require_once('relationship_rdf.module'); // use this for the import, even if it's not enabled yet
    relationship_rdf_read_from_source($source);
    relationship_flush_terms();
  }

  print (theme('page', $out));
}


/**
 * Impliment the requested database maintainence.
 * @see relationship_check_database()
 * @ingroup relationship_admin
 */
function relationship_database_maintainence_submit($form_id, $form_values){
  $nids = $form_values['nid'];
  if (is_array($nids)){
    drupal_set_message('Removing references to ' . count($nids) . ' objects..');
    foreach ($nids as $nid){
      relationship_resource_delete_statements($nid, TRUE);
    }
  }
  $invalid_statement_query = " (TRIM(subject) = '') OR (TRIM(predicate) = '') OR (TRIM(object) = '') ";
  $query = "DELETE FROM {relationship_statement} WHERE $invalid_statement_query";
  db_query($query);
  drupal_set_message(t('Cleaned up database'));
}


/**
 * @ingroup relationship_api
 * @{
 */

/**
 * Check that the given object is suitable for use as a predicate term.
 *
 * Fill in any gaps, and abstract the current properties into statements.
 * relationship_save_term() is the UI form that usually calls this.
 * This is an actin that actually does stuff, distinct from
 * relationship_term_is_valid() which is just a check. 
 * @param $term object definition
 * @return success
 * @ingroup relationship_api
 * @see relationship_delete_term()
 * @see predicates The predicate term object definition.
 */
function relationship_validate_term(& $edit){
  // A term object is really an abstract collection of statements.
  // Saving the statements collectively means saving the term.

  if (!$edit['name']){ // This is an API, more paranoia
    drupal_set_message('Cannot work with a term with no name identifier at all! Problem updating term. ' . print_r($edit, 1), 'error');
    return;
  }

  debug('validating term ' . $edit['name'], 2);

  if (!$edit['uri'])
  { // This is an API, more paranoia
    $edit['uri'] = $edit['name'];
    // Must have a uri now, should create a localhost id I guess
    debug(t('
			        Setting term URI to just "%uri" - this means that the term is 
			        basically only valid in the context of the current site. 
			        This is OK for made-up relationships, but means this bit of data
			        is harder to share with other servers, as it has no namespace.
			    ', array ('%uri' => $edit['uri'] )
    ), 1);
  }

  $edit['type'] = ($edit['type']) ? $edit['type'] : CLASS_PROPERTY;
  // This statement is what makes a predicate a predicate

  // check if there's an older version
  $old_term = relationship_get_term($edit['name']);
  if ($old_term && ($edit['uri'] != $old_term->uri)){
    // If they've changed the URI, it's annoying, as that's the UID for the term
    // Need to clear out everything that it used to know.
    // The update below will replace it with new values. Hopefully it won't break anything.
    relationship_delete_term($old_term->uri);
  }

  #debug_pre($term);
  debug("Generating statements from term definition", 3);
  $edit['statements'] = relationship_array2statements($edit, $edit['uri'], relationship_get_property_map($edit['type']), NULL, 'From Term Definition Validation');

  // For convenience (but not efficiency) I'll try to use shortnames when saving statements
  // about my own ontologies. This will have side-effects later on.
  foreach ($edit['statements'] as $i => $s){
    relationship_validate_statement($edit['statements'][$i]);
  }

  debug("<h2>Validated term " . $edit['name'] . "</h2>", 1);
  #debug_pre($edit);

  return $edit;
}

/**
 * Returns just true or false if the predicate term has no troubles.
 */
function relationship_term_is_valid($term){
  if (!$term->name){
    $problem = 'Cannot work with a term with no name identifier at all.';
  }
  if (!$term->uri){
    $problem = 'Term has no URI.';
  }
  if (!$term->label){
    $problem = 'Term has no label.';
  }

  if($problem){
    drupal_set_message($problem . '<pre>'.print_r($term, 1).'</pre>', 'error');
    return FALSE;
  }
  return TRUE;
}

/**
 * Return a list describing all possible properties that may apply to a type.
 * 
 * This list may be modified by plugins adding extra logic and making
 * appropriate term definitions claiming their new characteristic has 'domain'
 * of 'property' or 'ObjectProperty'.
 *
 * As this list is pretty stable, it's cached. The cache should be flushed
 * whenever a term is modified, and it'll be rebuilt if empty.
 *
 * @param $flush if set, dump the cache and force a re-read, re-calculation of the data
 * @return assoc array of available characteristic definitions. 
 *         Indexed by the characteristic shortname.
 *         The definition should include the label and the URI of the characteristic
 * @see apply_characteristics.inc
 * 
 */
function relationship_get_all_characteristics($type = CLASS_PROPERTY, $flush = FALSE){
// TODO deprecate this, its a waste of time?
// At least do some metrics on how often it's needed - is caching right?'
  $characteristic_cache_key = "relationship_characteristics";
  if($flush){
    cache_clear_all($characteristic_cache_key);
  }

//  $cached = cache_get($characteristic_cache_key);

  if ($cached->data){
    return unserialize($cached->data);
  }
  debug("Retrieving all properties that may apply to items of class $type", 2);

  // Characteristics are those predicates whose 'domain' 
  // (things it applies to) is 'property'. 
  //$characteristic_list = relationship_query( NULL, PREDICATE_DOMAIN, $type, 'subject' );
  //debug_pre($characteristic_list, 0);
  //
  // should be equivalent to:
  $characteristic_list = relationship_query($type, PREDICATE_CAN_HAVE_PROPERTY, NULL, 'object');
  debug_pre($characteristic_list, 3);

  // Thanks to inference, this list returns duplicates, due to the keying on full statement triples.
  // there may be more than one reason why the statement is returned, we've retrieved them ALL
  // Refactor this list to use all longnames
  $characteristics = array ();
  foreach ($characteristic_list as $k => $s){
    $s = relationship_get_longname($s);
    $characteristics[$s] = relationship_get_term($s);
  }
  cache_set($characteristic_cache_key, serialize($characteristics));
  return $characteristics;
}

/**
 * @}
 */

  

/**
 * Import a structured array of info 
 * as a set of statements.
 *
 * When given an optional $import_template, certain terms will be mapped
 * from the structure array keys to statement predicates.
 * as in @see relationship_array2statements()
 *
 * This will also CLEAR OUT any previous statements made about
 * the subjects of the import.
 *
 * The new statements will be indexed on the uri or the identifier of the 
 * given set of data - or the $subject if given. In that order of precedence.
 * If not told otherwise, this assumes that the import statements define a TERM (type:CLASS_PROPERTY)
 *
 * @see relationship_import_definition() for the singular version of this action
 * @param $defs an array of definition arrays.
 */
function relationship_import_definitions($defs){
  if (!$defs){
    drupal_set_message(t('No term definitions to import, an empty array passed in when trying to import definitions from a module ontology.'), 'error');
    return;
  }

  foreach ($defs as $subject => $subject_data){
    if ($subject_data){
      if (!$subject_data['name']){
        $subject_data['name'] = $subject;
      }
      if ($term = relationship_validate_term($subject_data)){
        // validate has figured it out into statements
        relationship_replace_resource_statements($term);
        drupal_set_message(t('Imported a set of statements defining what "%term" means.', array (
          '%term' => $subject_data['name']
        )));
      }
      else{
        drupal_set_message(t('Trouble importing the definition of "%subject" - apparently the supplied data did not return as a valid term.', array (
          '%subject' => $subject
        )));
      }
    }
  }
  // Usually need to flush the predicate terms cache right about now, we have some new terms.
}