<?php
// $Id: administer_relationships.inc,v 1.46 2006/10/14 15:52:58 dman Exp $

/**
 * @file
 *  
 * Supports the node-specific APIs and hooks for relationship
 * Renders the relationship table form,  validates and saves links against nodes.
 *
 * A node can now have any number of 'statements' attached to it.
 * A 'statement' is an RDF 'triple' relating one resource with another.
 * EG,  the "next" resource from page one is page two.
 *
 * Simple and complex stuctures can be built from a collection of such
 * relationships.
 * 
 * Primary statements behave as if they are attached directly to the
 * subject - the node. They are exposed for editing in the node edit form
 * and should be saved alongside it. Secondary statements are inferred
 * from logical patterns,  and although they may get stored in the node
 * as $node->relationships during processing and rendering,
 * they should not be saved unless you want every logical relationship 
 * to become explicit.
 *
 * This file provides the actual relationship editing UI form, and not much
 * else. To this end it uses the range_selectors.inc and combobox.js
 * to provide the dynamic UI.
 *
 * @package coders
 * @author Dan Morrison http://coders.co.nz/
 * @version $Id: administer_relationships.inc,v 1.46 2006/10/14 15:52:58 dman Exp $
 */


// we need to filter out some of the predicates, both for security
// and to keep the list managable.
// These constants describe what we show

/**
 * Intrinsic values are those predicates that have a direct mapping with
 * Drupal internals. Need to control access to this group.
 * An intrinsic property is one that has a drupalEquiv mapping value set.
 */
define('RELATIONSHIP_FILTER_INTRINSIC_VALUES',1);

/**
 * Trivial values are predicates that have an importance/weight less than zero.
 * This includes predicates that are only ever implied. They should be not
 * show up on user edit forms
 */
define('RELATIONSHIP_FILTER_TRIVIAL_VALUES',2);

/** @defgroup admin 
 *  
 * Administrator Interface
 * Hooks and functions for administering node relationships.
 * @{
 */





/**
 * Generate a form for displaying and selecting relationships applying to a
 * node.
 *
 * The form is supported by some DHTML on the client.
 * @see coders_widgets.module
 *
 * When adding in instance of this sub-form to a form, also add  the following
 * to instruct it to save its own data:
 * $form['#submit'] = array('relationship_form_submit'=>array(),);
 *
 * 
 * @param $node object - the resource
 * @param $predicates optional list of predicates to be used in the relationship drop-down
 * If not given, the list will be constructed according to the domain rules for
 * this object type, and filtered according to visibility settings.
 * @param $objects optional list of urls to be used in the relationship object drop-down
 *
 * @return HTML table form for adding metadata statements
 *
 * @TODO significant rewrite. Deal with nids 
 */
function relationship_form(&$resource, $predicates=NULL) {

  debug("Constructing relationship form",2);
  // Enable combo box functionality
  if(module_exist('coders_widgets')){
    coders_widgets_init('combobox');
    coders_widgets_init('paired_selects');
    coders_widgets_init('editable_hidden_fields');
  } else {
    drupal_set_message("This form requires coders_widgets.module to be enabled to do its stuff nicely.");
  }

  if (!isset($resource->statements)) { 
    if ($resource->nid) {
      // fetch existing statements
      // Should have already been loaded, but..
      debug("shouldn't get here, rendering statements for a node that should have been initialized but wasn't. (maybe just in preview)",2);
      $resource->statements = relationship_query( relationship_get_uri($resource) );
    }
    else {
      // new, empty resource
      debug("New resource, seems like there was no info given ".print_r($resource,1),2);
      if(! $resource){ $resource=(object)$resource; }
      $resource->statements = relationship_array2statements($resource,'new_'.$resource->type);
    }
  }

  # drupal_set_message("Displaying relationship statements <pre>".print_r($resource, 1) ."</pre>");
  debug_pre($resource,2);

  $relationship_terms = relationship_get_terms();
  

  // Initialize javascript selectbox updater
  include_once('range_selectors.inc');


  // Figure what predicates are appropriate to display and edit.
  // The FILTERs and individual relationships defs should tell us what's editable and what's not.
  $options_filter = 0;

  $show_trivial = variable_get('relationship_show_advanced_terms',FALSE);
  $show_intrinsic = variable_get('relationship_show_intrinsic_terms',FALSE);
  $can_edit_intrinsic = user_access('edit intrinsic properties');

  if(! ( $show_intrinsic && $can_edit_intrinsic )){
    $options_filter = RELATIONSHIP_FILTER_INTRINSIC_VALUES;
  }

  if(! $predicates){
    // This lookup should fetch not just explicitly defined
    // properties, but also others according to the lookup rules in 
    // relationship_lookup_CanHaveProperty() eg, properties inheirited by virtue of class.
    // todo - relationship_lookup_CanHaveProperty() needs work
    // @see relationship_get_all_characteristics() - a caching version of this

    debug("Retrieving all predicate relationships this resource type ({$resource->type}) MAY possible have",2);

    $predicate_list = relationship_query($resource->type, PREDICATE_CAN_HAVE_PROPERTY, NULL, 'object');
    $predicates = array_map(relationship_get_term, $predicate_list);

    debug('Predicates which MAY be appropriate to display on this '.$resource->type.' form are:<pre>'.print_r(array_values($predicate_list),1).'</pre>' ,2);
  
    $options_filter = ($options_filter | RELATIONSHIP_FILTER_TRIVIAL_VALUES);
  } else {
    debug('This form has been handed a list of which predicates to expose:' ,1);
    debug_pre(array_keys($predicates) ,1);
  }

  // flatten the definitions into a (restricted) option list
  $predicate_list = relationship_terms_as_options($predicates,$options_filter);

  debug('FILTERED Predicates appropriate to display on this '.$resource->type.' form, according to rules and inheiritance are:<pre>'.print_r($predicate_list,1).'</pre>' ,1);
  
  if(! $predicate_list){
    // This can only happen with a broken install or bootstrap.
    drupal_set_message(t("There are no known predicates (relationships or properties) in the local schema that are appropriate to apply to this resource. Create or import some predicate terms, then apply them to items of this type. The relationship selectbox will be empty."),'error'); 
  }


  // List all existing data in a table here
  // There are a few extra ids and things in the layout
  // for javascript widgets to do their job.
  
  $rows = array(
    '#type' => 'formtable',
    '#id'   => 'metadata_entry_table',
  );
  $rows['header'] = array('#type' => 'tr', 'this', 'relates to', '', array('#value'=>'that','#width'=>'100%'), '');

  $rows['header'] = array('#value' => '<thead><tr><th>this</th><th>relates to</th><th></th><th width="100%">that</th><th></th></tr></thead>');


  if(is_array($resource->statements)){
    relationship_get_titles($resource->statements);

    $a_table_cell = array('#type' => 'td');
 
    foreach ($resource->statements as $i => $statement) {
      // under 4.7 it seems I have to number all my multiple fields, can't just use [] no more
      // index on the statement ID if known, else use the current key, which will be a unique (long) hash
      $delta = ($statement['rsid'])?$statement['rsid']:check_name($i);
      
      debug("(possibly) Creating form row for statement $i",2);

      // TODO this routine needs to be refactored - it was constructed sorta ad-hoc
      // logic is currently broken down into the tiniest pieces, to really lay out whats happening

      $can_edit = TRUE;

      if(! $predicate_list[$statement['predicate']]){
        debug("This type of statement '".$statement['predicate']."' isn't on the list of things that this user can mess with. ".join(',',array_keys($predicate_list))." Not exposing it for editing on this form.",2);
        $can_edit = FALSE;
      }

      $is_trivial = ($statement['depth'] < 0);
      $is_explicit = ($statement['depth'] == 0);
      $is_deduced = ($statement['depth'] > 0);

      if( $is_trivial && ! $show_trivial){continue;}

      $row = array(
        '#type' => 'tr',
        'subject_td'   => $a_table_cell,
        'predicate_td' => $a_table_cell,
        'change_td'    => $a_table_cell,
        'object_td'    => $a_table_cell,
        'action_td'    => $a_table_cell,
      );      


      // This statement should be deleted, remember this through the previews
      $row['action_td']['delete'] = array(
        '#type' => 'hidden',
        '#value' => $statement['delete'],
        '#parents' => array('statements',$delta,'delete'),
      );
      if($statement['delete']){
        debug('remember to delete this '.$statement['predicate'].' statement',3);
        $row['#class'] = 'delete';
        $row['#attributes']['style'] = 'color:#CCCCCC; text-decoration:line-through';
        $row['#attributes']['title'] = 'this statement will be deleted on save';
        $is_explicit = FALSE; // disable further editing
      }

      // TODO One day I should put fine-tuned authentication here.
      // It's reasonable to set permissions on a predicate-by-predicate basis.
      
      $term_def = $relationship_terms[$statement['predicate']];
      $is_intrinsic = $term_def->drupalEquivalent;

      if( $is_intrinsic && ! $show_intrinsic){continue;}

      if( $is_intrinsic && ! $can_edit_intrinsic ){ $can_edit = FALSE; debug($statement['predicate']." is intrinsic, and you can't edit them directly",2);}
      if( $is_deduced ){ $can_edit = FALSE; }

      // if a predicate has no defined range (eg 'Node' or 'Resource')
      // treat it as a literal ... UNLESS it already looks like an URL
      // these are just guesses to make up for incomplete ontologies.
      // Having no defined range is bad.
      $is_literal = (($term_def->range == CLASS_LITERAL) || (!$term_def->range && !is_uri($statement['object']) && !is_numeric($statement['object'])) );


      // Display subject only if it's different than expected (an alias)
      $row['subject_td'][] = ($resource->uri && ($statement['subject'] != $resource->uri))
                      ?  "<span class='subject'>{$statement['subject']}</span>" 
                      : '';

      $row['predicate_td'][] = array(
        '#type' => 'markup',
        '#value' => '<span class="predicate" title="'.$term_def->name."\n".check_plain($term_def->description).'">'. (($term_def)? $term_def->label : $statement['predicate']) . '</span>'
      );

      if($can_edit){
        // Statement is partially or fully editable
        
        // need hidden fields
        $row['subject_td'][] = array(
          '#type' => 'hidden',
          '#value' => $statement['subject'],
          '#parents' => array('statements',$delta,'subject'),
        );

        $row['predicate_td']['predicates'][] = array(
          '#type' => 'hidden',
          '#value' => $statement['predicate'],
          '#parents' => array('statements',$delta,'predicate'),
        );

        $row['object_td']['depths'][] = array(
          '#type' => 'hidden',
          '#value' => $statement['depth'],
          '#parents' => array('statements',$delta,'depth'),
        );
        $row['object_td']['rsid'][] = array(
          '#type' => 'hidden',
          '#value' => $statement['rsid'],
          '#parents' => array('statements',$delta,'rsid'),
        );

				// The HTML id of a changeable (text) form element - used by widget code
        $target_object_id = "hidden_object_${delta}";
        $object_element = array(
          '#type' => 'hidden',
          '#id' => $target_object_id,
          '#attributes' => array('class'=>'changeable'),
          '#value' => $statement['object'],
          '#parents' => array('statements',$delta,'object'),
        );

        $link_atts = array('id' => 'exposed_object_'.$delta);
        
        // can rename the title
        if ($is_explicit ){
          if(!$is_literal){
            // Only explicit links can be titled
            // Only save titles that are different from the default. 
            // No need to fill up the info with redundancies, and they have a disturbing habit
            // of getting out of date.
            $title = ($statement['title'] != $statement['object'])?$statement['title']:'';

            $title_element = array(
              '#type' => 'hidden',
              '#attributes' => array(
                'id' => 'hidden_'.$delta, 
                'class'=>'changeable'
              ),
              '#title' => 'Label:',
              '#value' => $title,
              '#parents' => array('statements',$delta,'title'),
            );

            $row['object_td']['titles'][$delta] = $title_element;

            $link_atts['id'] = 'exposed_object_'.$delta;
          } else {
            $object_element['#class'] = 'changeable';
            $object_element['#title'] = 'Value:';
          }
        } 

        $row['object_td']['objects'][] = $object_element;
#debug_pre($statement);
        $row['object_td']['link']['#value'] = relationship_make_link( $statement, $link_atts );

        
        if ($is_explicit && (! $is_literal)  && module_exist('coders_widgets') ){
          // Trigger possible widget script
          $row['change_td']['#value'] = '<input type="button" value=">" '.
            ' onClick="change_select_range_to(\''.$target_object_id.'\',\''.$statement['predicate'].'\'); init_combo_box_named(\''.$target_object_id.'\') ; remove_element_named(\'exposed_object_'.$delta.'\') " title="Choose from other valid values for this property" />';
        }


        if ($is_explicit && module_exist('coders_widgets') ){
          // Only Explicit statements can be deleted
          // Writing button by hand as the drupal render is too bulky to fit in a table
          // This primarily works as javascript, but falls back as a submit action on noscript
          $row['action_td']['del']['#value'] = '<input type="submit" value="Del" name="edit[statements]['.$delta.'][delete]" '.
            ' onClick="if (typeof(delete_row_from_edit_table)==\'function\') {return delete_row_from_edit_table(this,this.name);}" />';
        } else {
          if ($is_trivial ) {
              $row['action_td']['trivial']['#value'] = "<small title='This actual statement is derived from data stored internally or elsewhere '>implicit</small>";
          }
        }

      } else {
        // can't edit this
        debug("You are not allowed to modify ".$statement['predicate']." on this object. intrinsic:$is_intrinsic deduced:$is_deduced",2);

        $row['object_td']['link']['#value'] = relationship_make_link( $statement );

        if( $is_deduced ){
          $row['action_td']['derived']['#value'] = "<small title='This actual statement is in another page, or inferred due to the logical structure of the documents. It cannot be directly edited from this page. \n\n".check_plain($statement['why'])." (".$statement['depth']." levels of inferrance)'>... by implication</small>";

        }
        if( $is_intrinsic ){
          $row['action_td']['intrinsic']['#value'] = "<small title='This value is handled by the Drupal internals, or another module.'>built-in</small>";
        }
          // And can't do much more than that with this statement
      }
      
      $rows[] = $row;
     
    } // each statement
  } // has existing statements


  // Done rendering all known statements
  // now,
  // Append a new editable row for a new statement 


  $predicate_list = array_merge(array(''=>'...'), $predicate_list ); //insert blank first field 
  // didja know that a blank option is apparently invalid XHTML?

  // Inline a buch of javascript configuration script
  $rows[]=array('#value' => range_selectors($predicates, 'predicates][', 'objects][' ) );
  
  // Don't put anything in the selector yet?
  $object_list = array('- Choose a relationship first'); 
  
  // the new row is constructed here  
  $rows[] = _relationship_statement_fields( $predicate_list, $object_list, $resource->uri );
  
  $form = array($rows);

  $group =  array(
    '#type'   => 'fieldset',
    '#title'  => t('Relationships and Metadata'),
    '#validate' => array('relationship_form_validate'=>array()),
    # '#submit' => array('relationship_form_submit'=>array()),
  );

  $group['relationships']= $form;
  
  $group['relationships']['help']['#value'] = t("<p>
      For more information about the available relationships, see %help_link.
      </p>
    ", array(
      '%help_link' => l(t('Relationship Help (Predicates)'),'admin/help/relationship/predicates'),
      '%resource_type' => $resource->type,
      '%class_def_link' => l("'".$resource->type."' definition",'admin/relationship/term/edit/'.$resource->type),
     )
  );

  return $group;
}


/**
 * Check the submission of any relationship table
 * 
 * unimplimented
 */
function relationship_form_validate($form) {
  debug('Validating relationship form',2 );
//  $resource = (object)array();
//  $resource = relationship_read_form($resource);
//  debug_pre($resource);
//  debug('Validating relationship form table input '.print_r($form, 1),0 );
}


/**
 * A resource table has been submitted. Save the data.
 * Often the tables will be incomplete, containing only the user-editable subset
 * of the full data. So when saving statments, the database entries are updated,
 * not replaced.
 */
function relationship_form_submit($form_id, $form_values, $subject='') {
  #dsm("Submitted some relationships as part of $form_id  ($subject). ");

  $resource = (object)$form_values;
  relationship_read_form($resource);

  # Not a full resource save, just save the statements
  relationship_update_resource_statements($resource, $subject);
}

/**
 * Convert a list of term def objects into an array
 * suitable for using in a select box.
 *
 * They will be indexed either of shortname (name) or longname (uri) depending on global settings
 *
 * Checks that only resources of type 'property' show up,
 * as relationship_lookup_CanHaveProperty() is currently a bit inaccurate
 * 
 * @param The current list or subset of predicate terms definitions.
 * @param $filter A boolean combination of options indicating which of the
 *        available terms NOT to show. 
 *        By default both filters are on - you must choose to turn them off.
 * @see RELATIONSHIP_FILTER_INTRINSIC_VALUES
 * @see RELATIONSHIP_FILTER_TRIVIAL_VALUES
 * @see drupal_get_terms()
 */
function relationship_terms_as_options($terms , $filter = 3 ) {
  if(! is_array($terms)){
    trigger_error("Invalid arg '$terms' to relationship_terms_as_options() ",E_USER_WARNING);
    return array();
  }
  $options = array();

  // index this list on long or short names
  $relationship_long_names = variable_get('relationship_long_names', FALSE);
  
  foreach($terms as $i => $term){
    debug("Filtering? $term->type $term->name weighted $term->weight , built-in? $term->drupalEquivalent ",3);
    // Skip the low-level terms - they are too trivial to list
    if(( $filter & RELATIONSHIP_FILTER_TRIVIAL_VALUES ) && ($term->weight < 0)){ continue; }

    // Do not list built-in properties unless given special permission
    if(( $filter & RELATIONSHIP_FILTER_INTRINSIC_VALUES ) && ($term->drupalEquivalent)){ continue; }

    if($term->type != CLASS_PROPERTY ){ continue; }

    // @TODO greater granularity of access can be applied here
    // else
    $indexid = ($relationship_long_names)?$term->uri:$term->name;
    $options[$indexid] = $term->label;
  }
  return $options;
}

/**
 * Read form values and absorb them into the given resource object as
 * 'statements'
 * 
 * It does not actually absorb the values back into the object - it just
 * interprets the form fields.
 * This bypasses the form API, as it may heve to find fields added on- the- fly
 * at the client side.
 * @param $resource - an object, passed by ref that may be updated
 * @param $subject string optional, gets set as the 'subject' of any statements
 * if the subject is otherwise undefined. Should be the $node->nid or the
 * $resiures->uri or $term->identifier or whatever
 */
function relationship_read_form(&$resource,$subject=''){
  debug("Reading submitted form values back into statements. Posted values are:",2);
  debug_pre($_REQUEST['edit'],3);
  debug_pre($_REQUEST['edit']['statements'],2);
  
  // remove any unused blanks
  foreach ((array)$resource->statements as $ix => $statement) {
    if(!$statement['predicate'] || ! $statement['object']){
     unset($resource->statements[$ix]); 
    }
  }

  // form api doesn't recognise any fields added by the client (for security reasons) 
  // but I need to bypass that . Read directly from the submission
  $submitted_statements = $_REQUEST['edit']['statements'];
  
  if (is_array($submitted_statements) ) {
    foreach ($submitted_statements as $ix => $statement) {
      array_map ('trim', $statement); 
      if ($statement && $statement['predicate'] && $statement['object']) {
        $statement['subject'] = ($statement['subject'])? $statement['subject'] : $subject;
        $statement['why']     = ($statement['why'])?$statement['why']:'user/'.$GLOBALS['user']->uid;
        // It's possible a 'rsid' number or a 'delete' flag will be sent through here also. Keep it.

        // re-index this item
        unset($resource->statements[$ix]);
        $resource->statements[ _relationship_hash_key($statement) ] = $statement;
      }
      
      // Do NOT want form API to find this now I've got it.'
      unset($_REQUEST['edit']['statements'][$ix]);
      unset($_POST['edit']['statements'][$ix]);
    }
  }
  // init title if needed
  relationship_get_titles($resource->statements, TRUE);

  debug("Finished reading form. statements are now:<pre>".print_r($resource->statements,1)."</pre>",2);
  return $resource;
}


/**
 * Ensure a sting is able to be used as an XML, CSS or Javascript ID.
 * Basically strip out all non-alpha-numerics
 * INCLUDING '-' which would break in javascript.
 * http://www.w3.org/TR/REC-xml/#NT-Name
 * @see form_clean_id() - which should have done this
 */
function check_name($name){
  return preg_replace('|[^a-zA-Z0-9_]+|','_',$name);
}


/**
 * Import a resources statements into top-level properties.
 * 
 * A more careful version of relationship_statements2array()
 * as it checks for authority to do so for each possibly low-level, intrinsic
 * property.
 * 
 * This should usually be called right after relationship_read_form()
 * 
 * @param $resource object. Expected to contain an array of ->statements which
 * are to be absorbed.
 */
function relationship_merge_statements_into_resource(&$resource) {

  // If given, allow relationship statements to be absorbed over top of
  // intrinsic node properties. This gives low-level access to the guts of a
  // node (probably a bad thing)
  $data_cloud = relationship_statements2array($resource->statements );
  foreach($data_cloud as $predicate => $object){
    if( isset($resource->$predicate) &&  ( $node->$predicate != $object) ){

      // Filter for user-access restrictions. Don't want to expose Drupal guts randomly
      if( user_access('edit intrinsic properties') ){
        debug("Over-writing node $predicate = {$node->$predicate} with $object ",2);
        $resource->$predicate = $object;
      } else {
        drupal_set_message(t('You don\'t have permissions to modify the %predicate directly on this resource, but thanks for trying!',array('%predicate'=>$predicate) ));
      }
    }
  }
  return $resource;
}





/**
 * Generate form elements for defining relationships.
 *
 * subject (current node, implied from context)
 * predicate (relationship term)
 * object (freeform,  probably a node id or URL)
 *
 * Return as an array so the respective parts can be
 * dealt with separately (put into a table row)
 *
 * Helper for node_form
 *
 * @param $predicates optional list of predicates to be used in the relationship drop-down
 *                    if not defined, ALL things of type 'Property' will be listed
 * @param $objects optional list of titles=>uris to be used in the relationship object drop-down
 * @returns an array to be put into a table row.
 * @todo test noscript
 */
function _relationship_statement_fields($predicate_list=NULL, $object_list=NULL, $subject='') {

  // fudging the parents array fools form_api into constructing the right name, id & data structure.
  // This row must be set with a null value (it's always a new entry) or forms api re-populates it
  // by being clever, which we don't need.
  $predicate_selector = array(
    '#type' => 'select',
    //'#target' => 'object_selector',
    '#target' => 'edit-statements-0-object',
    '#title' => '',
    '#default_value' => '',
    '#value' => '',
    '#options' => $predicate_list,
    '#description' => NULL,
    '#parents' => array('statements','0','predicate'),
  );

  $object_selector = array(
    '#type' => 'select',
    // '#id' => 'object_selector',
    '#title' => '',
    '#default_value' => '',
    '#value' => '',
    '#options' => $object_list,
    '#description' => NULL,
    '#attributes' => array('class' => 'combo'),
    '#parents' => array('statements','0','object'),
  );
  $hidden_subject = array(
    '#type' => 'hidden',
    '#id' => 'hidden_subject',
    '#default_value' => $subject,
    '#parents' => array('statements','0','subject'),
  );
 
 // Construct a row, rather laboriously
  $row = array(
    '#type' => 'tr' ,  
    'subject_td'   => array(
      '#type'        =>'td', 
      'data'         => array('#value' => 'this '),
      'subjects'     => array('' => $hidden_subject) ,
    ),
    'predicate_td' => array(
      '#type'        =>'td', 
      '#attributes'  => array('colspan'=>'2'),
      'predicates'   => $predicate_selector ,
    ),
    'object_td'    => array(
      '#type'        => 'td',
      'objects'      => $object_selector ,
    ),
  );
  
  if(module_exist('coders_widgets')){
    // add DHTML form enhancements
    coders_widgets_init('dynamic_table_rows');
    $row['predicate_td']['predicates']['#type'] = 'paired_select';
    $row['action_td'] = array(
      '#type'  => 'td',
      '#value'   =>'<input type="button" value="New" name="new_statement" id="new_button" '.
      'onClick="if (typeof(add_row_to_edit_table)==\'function\')'.
      '{return add_row_to_edit_table(this);}'.
      'else{alert(\'no dhtml function available\');return false;}" />',
    );
  }
  return $row;
}



/**
 * A query lookup callback, returns all predicates that may be available to a type.
 *
 * All the properties a subject can have!
 * This includes a logical lookup not implied by the ontology.
 * These callback lookups are where we CAN put business logic that we can't
 * yet express in statements.
 *
 * In addition to those properties added explicitly in the datastore, we should return 
 * - all properties from any associated ontologies, 
 * - all properties that may apply to the superclass (node)
 * - all wildcard properties that have NO domain defined.
 *
 * @TODO check that members of ontologies really are 'properties'
 *       I'm getting a few false properties due to classes, non-node properties
 *       and ontologies being inside ontologies too.
 *       I don't really want EVERYTHING in an ontology.
 * @ingroup ontologies
 */
function relationship_lookup_CanHaveProperty($subject){
  if(!$subject){$subject = CLASS_RESOURCE;}

  debug("What properties is a '$subject' allowed to have?",2);
  $statements = array();
  
  // Lookup possible predicate groupings - used when importing
  // content type definitions from custom ontology sets
  $possible_ontologies = relationship_query($subject,PREDICATE_CAN_USE_ONTOLOGY,NULL,'object');
  foreach($possible_ontologies as $ontology_name){
    // for all ontologies this class participates in...
    $ontology_statements = relationship_query(NULL, PREDICATE_IS_PART_OF, $ontology_name);

    $ontology_statements = relationship_query(NULL, PREDICATE_IS_PART_OF, relationship_get_longname($ontology_name));
    
    // each member of that ontology
    foreach($ontology_statements as $statement){
      // Note - that was a reverse lookup, so we use the 'subject' as our result
      // adjust all the derived statements to refer to the current subject
      $s = relationship_statement( $subject, PREDICATE_CAN_HAVE_PROPERTY , relationship_get_shortname( $statement['subject'] ) );
      $s['depth'] = (int)$statement['depth'] + 1;
      $s['why'] = $statement['why']."; in ontology $ontology_name, which is associated with ".$subject;
      $statements[_relationship_hash_key($s)] = $s;
    }
  }
 
  // all properties that may apply to the superclasses (node,Resource)
  if($subject != CLASS_CLASS){ // avoid loop
  // CLASS_RESOURCE
    $superclasses = relationship_query($subject, PREDICATE_SUBCLASS_OF,NULL,'object');
    // if(! $superclasses){debug("odd, this $subject doesn't seem to be an anything - it has no superclass");}
    foreach($superclasses as $superclass){
      debug($subject." is a subclass of '$superclass', so it can have its properties also",2);

      // TODO should this be recursive?
      $superclass_statements = relationship_query($superclass, PREDICATE_CAN_HAVE_PROPERTY, NULL);
      // adjust all the derived statements to refer to the current subject
      foreach($superclass_statements as $statement){
        if(!$statement['object']){debug('Query returned a statement with no object. this is not right. This code should not have been reached');}

        // $s = relationship_statement( $subject, PREDICATE_CAN_HAVE_PROPERTY , relationship_get_shortname($statement['object']) );
        $s = relationship_statement( $subject, PREDICATE_CAN_HAVE_PROPERTY , $statement['object'] );
        $s['depth'] = (int)$statement['depth'] + 1;
        $s['why'] = $statement['why']."; in ontology $ontology_name, which is associated with class: ".superclass;
        $statements[_relationship_hash_key($s)] = $s;
      }
    }
  }

  debug("Properties a '$subject' may have by virtue of lookup or inheiritance: ".join(', ',array_keys($statements)) ,3);

  foreach(relationship_get_terms() as $term){
    // all wildcard properties that have NO domain defined.
    if($subject && $term->name && !$term->domain)
    {
      $s = relationship_statement( $subject, PREDICATE_CAN_HAVE_PROPERTY , $term->name );
      $s['depth'] =  RELATIONSHIP_IS_CALCULATED;
      $s['why'] = "$term->name is a predicate term with no domain limit";
      $statements[_relationship_hash_key($s)] = $s;
    } 
    else if($term->domain == CLASS_RESOURCE)
    { // This term is explicitly defined as applying to all resources 
      $s = relationship_statement( $subject, PREDICATE_CAN_HAVE_PROPERTY , $term->name );
      $s['depth'] =  RELATIONSHIP_IS_CALCULATED;
      $s['why'] = "$term->name is defined to be a property available to all resources";
      $statements[_relationship_hash_key($s)] = $s;
    }
  }

  $reasons = array_map(create_function('$s', 'return $s["why"];'), $statements);
  debug("Properties a $subject may have by virtue of lookup, inheiritance or wildcards (not counting direct statements): <li>".join("</li>\n<li>",array_keys($reasons))."</li>" ,3);

  return $statements;  
}

/**
 * Return a list of type statements - What is every item of a given 'type'?
 * 
 * As well as the explicit statements, some Drupal objects are retrived like
 * this, so you can ask "What's everything of type 'page'?" (fetch all pages)
 * or everything of type "user" - to return the user list.
 * Supports inheiritance through subclasses too!
 */
function relationship_lookup_type($subject,$query){
  $members = array();
  if($query['object'] && (! $query['subject']) ){
    // Someone's asking for all resources of a given type.
    // No problem, only if 'a' is of type 'B', and 'B' is a subclass of 'C',
    // then 'a' is implicitly also of type 'C' so someone looking for all 'C's
    // should be given 'a'
    //
    // Inheiritance.
    //
    $superclass = $query['object'];
    debug("Type lookup requires inheiritance support. What's Everything of type $superclass",2);
    $subclasses = relationship_query(NULL, PREDICATE_SUBCLASS_OF, $superclass,'subject');
    foreach($subclasses as $subclass){
      debug("'$subclass' is a subclass of '$superclass', so every resource of type '$subclass' should also be returned when asked for everything of type '$superclass' ",2);
      $submembers = relationship_query(NULL, PREDICATE_TYPE,$subclass); 
      $members = array_merge($members,$submembers);
    }
    
    // Also, return a list of nodes of a certain 'type' if asked.
    // These values are retained inside Drupal, not the datastore.
    $node_types = node_get_types();
    if($type_def = $node_types[$query['object']]){
      // Return all nodes of this type
      // Isn't there an API for this request?
      // copied from the snippets
      $sql = "SELECT title, nid FROM {node} WHERE type = '%s' ";
      $result = db_query($sql,$query['object']);

      while ($anode = db_fetch_object($result)) {
        $members[] = relationship_statement(
          'node/'.$anode->nid,
          PREDICATE_TYPE,
          $query['object'],
          $anode->title,
          -1,
          'Drupal Database defines node type'
        );
      }
      
    }
  }
  
  return $members;
}


/**
 * Vanilla data dump of the statements about a given resource
 */
function relationship_list_statements($resource) {
  if(! is_array($resource->statements)){return '';}

  // May optionally be hiding the built-in data and only show the relationship.module additions
  $show_trivial = variable_get('relationship_show_advanced_terms',FALSE);
  $show_intrinsic = variable_get('relationship_show_intrinsic_terms',FALSE);
  $relationship_terms = relationship_get_terms();
  
  foreach ($resource->statements as $statement) {
    $pred = $statement['predicate'];

    if(($pred == 'title') || ($pred == 'label')){
      // if the resource has a label or title defined in this statement
      // collection, display that as the identifier instead
      $label = $statement['object'];
    }

    // Filter out the data bits we don't want to see.
    $is_trivial = ($statement['depth'] < 0);
    if( $is_trivial && ! $show_trivial){continue;}
    $term_def = $relationship_terms[$pred];
    $is_intrinsic = $term_def->drupalEquivalent;
    if( $is_intrinsic && ! $show_intrinsic){continue;}

    if(is_array($pred)){debug_pre($statement);}

    if($term_def->{PREDICATE_RANGE} && ($term_def->{PREDICATE_RANGE} != CLASS_LITERAL)){
      // This object can probably be linked to
      $object_link = l( $statement['object'], variable_get('relationship_browser_root', 'relationship/view') ,NULL, 'edit[subject]='.urlencode($statement['object'])  );
    } else {
      $object_link = '';
    }
  
    $output .= "<dt>".$pred."</dt>\n";
    $output .= "<dd>".($object_link?$object_link:$statement['object'])."</dd>\n";
    
  }
  if(!$output){$output = t('<p>No public data available</p>');}
  else{$output = '<dl>'.$output.'</dl>';}
  return $output;
}



/**
 * List everything about everything. Page.
 *
 * Menu callback.
 * Admin function / diagnostic / scaffolding
 * Should really use pager
 */
function relationship_list_all() {

  drupal_set_title('Dumping all relationships in the database table');
  $output = 'Debug info only - this will get too big very quickly';
  
  $all_relationship_lookup = db_query('SELECT * FROM {relationship_statement} order by subject;');
  $rows = array();
  while ($statement = db_fetch_array($all_relationship_lookup) ) {
    $row = $statement;
    $row['subject'] = l( scrunch_url( $statement['subject'] ), 'relationship/view', NULL, 'edit[subject]='.urlencode($statement['subject']) );
    unset($row['title']);
    $rows[] = $row;
    $headers = array_keys($statement);
  }
  unset($headers['title']);
  $output .= theme('table', $headers  , $rows);

  print theme('page',$output);  
}


function relationship_choose_resource_form(){
  $form = array();
  $form['uri'] = array(
    '#type' => 'textfield',
  );
  $form['edit'] = array(
    '#type' => 'submit',
    '#value'=> 'Edit',
  );
  $form['view'] = array(
    '#type' => 'submit',
    '#value'=> t('View'),
  );

  return drupal_get_form('relationship_choose_resource_form', $form);
}

/**
 * Expose all statements about any subject.
 *
 * Menu callback.
 * This treats the given resource as a stand-alone collection of statements,
 * with no special meaning or treatment.
 *
 * The target subject can either be handed in as the remainder or the URL
 * or explicitly as a 'uri' form parameter.
 * This is a generic edit form for a resource. A resource is represented as
 * a collection of statements about the same thing.
 *
 * @see relationship_list_all()
 * @todo make this a saving edit form
 */
function relationship_data_page($op='View',$subject='') {

  $op = $_REQUEST['op'] ? $_REQUEST['op'] : $op ; // Form op overrides path op
  
  // absorb subject from any direction possible
	// I'll allow requests for both ?edit[subject]=thing and ?subject=thing
  // Or just appended to the request (for shortnames only)
  $edit = array_merge( $_REQUEST,($_REQUEST['edit'])?$_REQUEST['edit']:array());
  $subject = 
  	$subject ? $subject :
    	( $edit['uri'] ? $edit['uri'] : 
        ( $edit['identifier'] ? $edit['identifier'] : 
        	$edit['subject'] 
        )
      );
  
  $actions = array('View'=>'view','Edit'=>'edit','RDF'=>'rdf');
  foreach($actions as $label=>$command){
    $active = ($op==$label)?array('class'=>'active'):NULL;
    $actions[$label] = l( $label, 'relationship/'.$command, $active , 'edit[subject]='.urlencode($subject) );
  }
  $out .= theme_item_list($actions,NULL,'ul',array('class' => 'tabs primary'));


  if(!$subject){
    drupal_set_message(t('No subject defined. Need to call this page with an id or a URI as the <code>?uri=</code> argument'),'error');    

    print( theme('page', relationship_choose_resource_form() ));
    return;
  }
  
	$resource = relationship_get_resource_by_name($subject, NULL, TRUE);

  if(! $resource ){
    drupal_set_message("The resource labelled '$subject' is totally unknown to the relationship database. Creating a generic resource placeholder.");
    $resource = (object) array(
      PREDICATE_IDENTIFIER => $subject,
      PREDICATE_TYPE => CLASS_RESOURCE,
    );
  }

	if(! $resource->uri ){
	  // brand new
	  $resource->uri = $subject;
	}

  if($op == t('edit')){
    // load submission data (if any)
    // Note this merges existing info with submitted
    // TODO authentication here? see relationship_merge_statements
    $resource = (object)array_merge((array)$resource,$edit);

  }
  
  if($op == t('Submit')){
    drupal_set_message("Saving '$subject'");
    $resource = (object) relationship_statements2array( $edit['statements'], NULL, TRUE );

    if(! $resource->uri ){ // brand new
      $resource->uri = relationship_get_longname($subject);
    }
    relationship_resource_save($resource);
  }

  drupal_set_title("$op All info about '$subject'");

  $out .= t("
    Viewing the properties and relationships that apply to the 
    <b>%type</b> named '%name'. ",
    array(
      '%type' => $resource->type?$resource->type:CLASS_RESOURCE, 
      '%name' => relationship_get_display_label($resource) 
    )
  );

  if($op == t('edit')){
    $form = relationship_form($resource);
  
    $form['submit'] = array(
      '#type' =>   'submit',
      '#value'=>   t('Submit'),
    );
    $form['#redirect'] = FALSE;
    
    $out .=  drupal_get_form('relationship_resource_form', $form);
  } else {
    // Assume 'Viewing'
    $out .= relationship_list_statements($resource);
  }
  
  print( theme('page', $out ));
}


/*! 
 *  @}
 */

/**
 * This form may be called either in the admin/settings/content-type page or the
 * CCK configuration page.
 * It requires its own custom callback function too.
 * Alters the given form by reference, a subroutine of hook_form_alter
 * 
 * @param $subject the identifier of the currently edited content type;
 */
function relationship_content_type_settings($subject,&$form)
{
  if(!$subject){
    debug('Blank subject submitted to settings form. Probably not an error if setting up a brand new entry.',2);
  }
  else {
    $resource =  relationship_get_resource_by_name($subject,NULL,TRUE);
    $uri = relationship_get_longname($subject); // NODE_TYPE_PREFIX.$subject;
  }
  
  if(! $resource){
    drupal_set_message("The relationship definition database has never heard of the node type $subject before now.");
    $type_names = node_get_types();

    $resource = (object)array(
      PREDICATE_LABEL => $type_names[$subject],
    );
  }
  // ALWAYS enforce these facts for all content types.
  //
  // It's a node
  $s = relationship_statement($uri,PREDICATE_SUBCLASS_OF,CLASS_NODE,'',0,'configured via content type settings');
  $resource->statements[_relationship_hash_key($s)] = $s;
  //
  // It has an identifier
  $s = relationship_statement($uri,PREDICATE_NAME,$subject,'',0,'configured via content type settings');
  $resource->statements[_relationship_hash_key($s)] = $s;

  
  // The form is not currently ordered :(
  // Do so now in order to find where to place our new field
  $index = 0;
  foreach(element_children($form) as $id){
    $form[$id]['#weight'] = $index++;
    if( ($form[$id]['#type']=='submit') || ($id == 'buttons') ){
      // locate the bottom of the form in two different form contexts 
      $submit_index = $index;
    }
  }


  // The relationship form here can only set two useful properties:
  // which predicates and which ontologies this content type will be able to use.
  // TODO replace this setting with a class canhaveproperty lookup
  $predicate_list = array( PREDICATE_CAN_USE_ONTOLOGY, PREDICATE_CAN_HAVE_PROPERTY, PREDICATE_SUBCLASS_OF, PREDICATE_SAME_AS );
  $predicates = array_map(relationship_get_term, array_combine($predicate_list,$predicate_list));

  $form['relationship_block'] = relationship_form($resource, $predicates) ;
  // Position it at the bottom of the form.
  $form['relationship_block']['#weight'] = $submit_index-1;
  
  // Adjust just the help text of this form group
  $form['relationship_block']['relationships']['help']['#value'] = t("The relationships that may apply to this content type: ".$resource->label );

  // capture a custom submission function, so we can save our own data as part of someone elses form.
  // Need to unshift this onto the FRONT of the submit queue, as the main submit aborts with a goto
  $form['#submit'] = array_merge(
    array('relationship_form_submit' => array($subject)), 
    // array('relationship_content_type_form_submit' => array()), 
    (array) $form['#submit'] 
  );    


  return $form;
}


/**
 * Save the parameters set on the content type definition page
 * (CCK hook)
 * @see relationship_form_alter() for the other end of this process
 * deprecated for the generic version?
 */
function relationship_content_type_form_submit($form_id,$edit){
  #drupal_set_message("relationship_content_type_form_submit callback!".print_r($edit,1));

  $content_def = (object)$edit;

  $content_def = relationship_read_form($content_def);
  relationship_merge_statements_into_resource($content_def);

  $subject_uri = relationship_get_longname($content_def->type);

  relationship_update_resource_statements($content_def,$subject_uri);
}

?>
