<?php
/**
 * @file
 * Library to extend the 'object' selector to show only appropriate targets
 * If the 'author' property should only point to users then the 'range' of
 * 'author' is 'users' and someone attempting to set the 'author' should 
 * be presented with a list of users.
 * Today we'll do it by populating a select via javascript.
 * Options are to AJAX something in the background ... next
 */


/**
 * Returns javascript to enable the dynamic selectbox updates
 * 
 * Inlines big javascript arrays so all options are available at all times.
 * 
 * TODO one day, perhaps fetch these arrays via AJAX on request rather than
 * inlining them. Replace with either autocomplete or dynamic selects
 */
function range_selectors($predicate_list, $trigger_prefix, $target_prefix) {

  if(! is_array($predicate_list)){
    debug("Invalid arg '$predicate_list' to range_selectors() ");
    return;
  }

  $relevant_ranges = relationship_get_selector_ranges($predicate_list);

 # debug_pre(array("relevant ranges from ",array_keys($predicate_list)," are", $relevant_ranges),1);

  $relevant_ranges['default'] = array(CLASS_NODE=>t("All Nodes"), CLASS_USER=>t("All Users"), CLASS_VOCABULARY=>t("All Terms"));

  // The set of 'resources' is the set of all members of all classes
  $relevant_ranges[CLASS_RESOURCE] = array_unique ( array_merge($relevant_ranges[CLASS_RESOURCE], relationship_get_range_labels() ) );
  // can I plural these labels?

  // Create a javascript config array for this info 
  // ... may go on-demand later, but inline is easier
  // This defines the mapping between predicates and ranges
  
  $js = "\n// Listing which selection values (on the left) should populate which ranges (on the right)\n";
  $js .= "// Each selection may refer to more than one range, so it's initialized with an ARRAY of range names,\n";

  $js .= js_array_dump($relevant_ranges ,'selector_ranges' );

  // Now serialize the range values themselves
  // list ALL members of ALL (relevant) ranges - pretty big pretty quick
  $required_range_definitions = array();
  debug("Looking up all the members of the all the ranges we need to know on this screen ".join(', ',array_keys($relevant_ranges)), 2);
  foreach($relevant_ranges as $predicate => $ranges){
    if( is_array($ranges) ){
      #debug("the range of $predicate => ".print_r($ranges,1));
      foreach($ranges as $range_name => $range_label){
        debug("Looking up the members of the range (things of type) '$range_name' because it's a potential value for the '$predicate' field.",3);
        if(! isset($required_range_definitions[$range_name]) ){
          $required_range_definitions[$range_name] = relationship_get_range($range_name);
        }
      }
    }
  }
  
  $js .= "\n\n// Now define the contents of the ranges\n";
  $js .= js_array_dump($required_range_definitions ,'ranges' );


  $js .= "\nvar TRIGGER_PREFIX = '$trigger_prefix'; var TARGET_PREFIX = '$target_prefix';\n";
  $js = "<script>\n$js\n</script>";
  return $js;
}




/**
 * Return a list of the ids and labels of all defined content types.
 *
 * ... plus a few built-ins like 'users' and 'vocabularies'
 * All 'Classes', or types of 'Resources' that Drupal manages.
 *
 * Each of these classes define ranges - possible sets of objects.
 * The range names 'node/type/rangename' (eg 'node/type/story' or
 * 'node/type/flexinode- 2' ) are MADE UP to represent the respective content
 * types. The returned array is normally keyed 'identifier'=>'Human name'
 * (label) If the long_name flag is set, the key is 'longname'(uri) => 'Human
 * name' (label)
 */
function relationship_get_range_labels($longname = FALSE) {
  $range_names = array();

  $range_names[CLASS_NODE] = t("All Nodes");

  // All defined node content types (they are sorta classes)
  foreach(node_get_types() as $key => $label){
    $range_names[$key] = $label;
  }

  // and anything we know is a type of Class
  $statements = relationship_query(NULL, PREDICATE_TYPE, CLASS_CLASS);
  foreach($statements as $statement){
    $key = relationship_get_shortname($statement['subject']);
    $label = $key; //     $statement['title']?$statement['title']:($range_names[$key]?$range_names[$key]:$key);
    $range_names[$key] = $label;
  }
  # debug_pre(array('All range labels are',$range_names,$statements));

  $range_names[CLASS_USER] = "All Users";

  // and the labels of our vocabularies
  if(module_exist('taxonomy')){
    // If all known vocabularies are made available as ranges, it makes it easy to select their members
    $vocabs = array_map('relationship_get_shortname_property',taxonomy_get_vocabularies());
    foreach($vocabs as $name => $label){
      //$range_names['taxonomy/term/'.$vocabulary->vid'vocabulary/'.$key] = $label;
      $range_names[CLASS_VOCABULARY.'/'.$name] = $label;
    }
  }

  return $range_names;
}
 
 
 /**
  * Return the values in a given range, as named in relationship_get_range_labels().
  *
  * Using the custom 'type' lookup, the internal Drupal ranges
  * (All nodes of a given type, all terms in a given vocabulary) are also
  * returned here when requested by name.
  * Beware name conflicts.
  * @see relationship_lookup_type()
  * 
  *  Internal ranges are all resources of a given
  * 'type', like 'property' (list all properties), 'Class', or 'Ontology'. more?
  * 
  * @todo maybe cache here. This is done a lot on edit screens.
  * 
  * @return a list suitable for use in a drop-down or JS array.
  */
function relationship_get_range($range_name) {
  $range = array();
  if(is_uri($range_name)){
    $range_name = relationship_get_shortname($range_name);
  }
  debug("Finding all members of the range(class) '$range_name'",3);
  $bits = explode('/',$range_name);
  switch ($bits[0]) {
    case CLASS_LITERAL :
      // No range at all. 
      break;

    case CLASS_NODE :
      // All nodes
      // isn't there an API for this request?
      // copied from the snippets
      $sql = "SELECT title, nid FROM {node}";
      $result = db_query($sql,$bits[2]);
      while ($anode = db_fetch_object($result)) {
        #$range[$anode->nid] = substr($anode->title,0,40); // don't get too long'
        $range['node/'.$anode->nid] = substr($anode->title,0,40); // don't get too long'
      }
      break;
    
    case CLASS_VOCABULARY :
      if(module_exist('taxonomy')){
        $vocabs = array_map('relationship_get_shortname_property',taxonomy_get_vocabularies()) ;
  
        if($vocab = $bits[1]){
          // if defined, return just terms from given vocab, else return terms from all
          $vocabs = array($vocab => $vocabs[$vocab] );
        }
  
        foreach($vocabs as  $vid => $vocab_name ){
          $tree = taxonomy_get_tree($vid);
  
          // from _taxonomy_term_select()   
          if ($tree) {
            foreach ($tree as $term) {
              $range['taxonomy/term/'.$term->tid] = $term->name;
            }
          }
        }
      }
      break;
      
    case CLASS_USER :
      // no API for user list either? lifted from user.module
      // Users are identified by the path to their user profile
      $sql = 'SELECT uid, name FROM {users}  ';
      $result = db_query($sql);
      while ($auser = db_fetch_object($result)) {
        if($auser->name){
          $range['user/'.$auser->uid] = $auser->name;
        }
      }
      break;

    case CLASS_CLASS :
      // as well as everything literally called a class,
      // all locally defined content types and other resources are valid 'classes'
      $range = relationship_get_range_labels();
      break;

    default :
      // try to retrieve everything that is of the given 'type';
      debug("Getting the range called '".$bits[0]."' - searching for all resources of type '".$bits[0]."'" ,2 );

      $statements = relationship_query(NULL, PREDICATE_TYPE, $range_name);
      foreach($statements as $ix => $statement){
        $range[$statement['subject']] = $statement['title']?$statement['title']:relationship_get_shortname($statement['subject']) ;
        // The use of title here is a quirk from the lookup - it's not the object title, it's the subject title.
      }
   
  } // switch
 
  if(! $range && ($range_name != CLASS_LITERAL)){ 
    debug("no results for range '$range_name' when populating the array for the selectbox. This range is an empty set. Basically this means there is no possible valid value for this predicate. Odd.",2);
  }
 
  return $range;
}


/**
 * Lookup a list of predicates and return a list of the union of their
 * respective ranges.
 *
 * The list can be indexed on uri or shortname.
 * The content of the list can be redundant OR a handle on the predicate
 * definition object itself. If given, this will save a lookup.
 *
 * @todo deal with shortnames - currently
 * indexing on URIs instead
 * Returns an indexed array id=>label
 */
function relationship_get_selector_ranges($predicates) {
  $ranges = array(CLASS_RESOURCE=>array());

#debug("Finding what ranges are needed to apply to the predicates ".join(',',array_keys($predicates)) );  

  if(! is_array($predicates)){
    trigger_error("Invalid arg '$predicates' to relationship_get_selector_ranges() ",E_USER_WARNING);
    return array();
  }
  
  $relationship_long_names = variable_get('relationship_long_names', FALSE);

  foreach($predicates as $predicate => $def){
    // index on shortname or longname? 
    $indexid = ($relationship_long_names)?$def->uri:$def->name;

    if(is_object($def)){
      if($def->range){
        // ranges MAY be multiple. assume the range into a self-indexed array
        if(! is_array($def->range)){
          $r = array( $def->range );
          $def->range = array_combine( $r,$r ); 
        }
        $ranges[$indexid] = $def->range;  

				if(is_array($ranges[$predicate])){
	        $ranges[CLASS_RESOURCE] = array_merge($ranges[CLASS_RESOURCE],$ranges[$predicate]);
				}
      }
    } else {
      $ranges[$indexid] = relationship_get_properties($predicate,  PREDICATE_RANGE);
    }

    debug("The range of $predicate $label is ".( ($ranges[$indexid]) ? print_r($ranges[$indexid],1) : "unrestricted (hence assumed to be any resource URL)" ),2 );
  }
  return $ranges;
}

/**
 * Renders an array in a javascript way.
 *
 * This should be put into a script block, even better as an include.
 * @todo figure out caching for this.
 */
function js_array_dump($ar,$name = NULL) {

  if(is_array($ar)){
    $name = preg_replace('/[^\w]/','_',$name);
    if( $name ) {$out .= "\n\nvar $name = " ;}
    else {$padd='  ';}

    $out .= "$padd{\n";
    $vals = array();
    foreach($ar as $var => $val){
      if(is_array($val)){
        $vals[] = "'$var':". js_array_dump($val) ."\n" ;
      } else {
        $vals[] = "  '$var':'".check_plain($val)."' ";
      }
    }
    $out .= join(",$padd\n",$vals);
    $out .= "$padd\n}";
  }
  
  return $out;
  
}


// PHP4 backport
if (!function_exists('array_combine')) {
  function array_combine($keys, $values) {
    $keysValues = array();
    foreach($keys as $indexnum => $key)
      $keysValues[$key] = $values[$indexnum];
    return $keysValues;
  }
} 

?>
